"""
Tests for contacts API cyclone server.
"""

import json

import yaml

from twisted.internet.defer import inlineCallbacks, returnValue

from vumi.persist.txriak_manager import TxRiakManager
from vumi.tests.helpers import VumiTestCase
from vumi.tests.helpers import PersistenceHelper

from go_api.cyclone.helpers import AppHelper

from go_contacts.backends.riak import RiakContactsBackend, contact_to_dict
from go_contacts.server import ContactsApi


class ContactsApiTestMixin(object):
    def mk_api(self):
        raise NotImplementedError()

    def request(self, api, method, path, body=None, headers=None, auth=True):
        raise NotImplementedError()

    def create_contact(self, api, **contact_data):
        raise NotImplementedError()

    def get_contact(self, api, contact_key):
        raise NotImplementedError()

    def contact_exists(self, api, contact_key):
        raise NotImplementedError()

    EXPECTED_DATE_FORMAT = "%Y-%m-%d %H:%M:%S.%f"

    OWNER_ID = u"owner-1"

    CONTACT_FIELD_DEFAULTS = {
        u'$VERSION': 2,
        u'bbm_pin': None,
        u'dob': None,
        u'email_address': None,
        u'facebook_id': None,
        u'groups': [],
        u'gtalk_id': None,
        u'mxit_id': None,
        u'name': None,
        u'surname': None,
        u'twitter_handle': None,
        u'wechat_id': None,
        u'extra': {},
        u'subscription': {},
    }

    def assert_contact(self, contact, expected_partial):
        expected = self.CONTACT_FIELD_DEFAULTS.copy()
        expected.update(expected_partial)
        expected[u"user_account"] = self.OWNER_ID
        self.assertEqual(contact, expected)

    def assert_contact_response(self, response, expected_partial):
        expected = self.CONTACT_FIELD_DEFAULTS.copy()
        expected.update(expected_partial)
        expected[u"user_account"] = self.OWNER_ID
        data = response[1]
        if u"key" in data:
            expected.setdefault(u"key", data[u"key"])
            expected.setdefault(u"created_at", data[u"created_at"])
        self.assertEqual(response, (200, expected))
        return data

    @inlineCallbacks
    def test_get(self):
        api = self.mk_api()
        contact = yield self.create_contact(
            api, name=u"Bob", msisdn=u"+12345")

        contact_key = contact[u"key"]
        resp = yield self.request(
            api, "GET", "/contacts/" + contact_key)
        self.assert_contact_response(resp, {
            u'key': contact[u"key"],
            u'created_at': contact[u"created_at"],
            u'msisdn': u'+12345',
            u'name': u'Bob',
        })

    @inlineCallbacks
    def test_get_non_existent_contact(self):
        api = self.mk_api()
        (code, data) = yield self.request(api, "GET", "/contacts/bad-id")
        self.assertEqual(code, 404)
        self.assertEqual(data, {
            u"status_code": 404,
            u"reason": u"Contact 'bad-id' not found.",
        })

    @inlineCallbacks
    def test_create(self):
        api = self.mk_api()
        resp = yield self.request(
            api, "POST", "/contacts/", json.dumps({
                u"msisdn": u"+12345",
                u"name": u"Arthur",
                u"surname": u"of Camelot",
            }))
        contact = self.assert_contact_response(resp, {
            u"msisdn": u"+12345",
            u"name": u"Arthur",
            u"surname": u"of Camelot",
        })
        contact_key = contact[u"key"]

        stored_contact = yield self.get_contact(api, contact_key)
        self.assert_contact(stored_contact, contact)

    @inlineCallbacks
    def test_create_with_extras(self):
        api = self.mk_api()
        resp = yield self.request(
            api, "POST", "/contacts/", json.dumps({
                u"msisdn": u"+12345",
                u"name": u"Arthur",
                u"surname": u"of Camelot",
                u"extra": {
                    u"quest": u"Grail",
                    u"government": u"monarchy",
                },
            }))
        contact = self.assert_contact_response(resp, {
            u"msisdn": u"+12345",
            u"name": u"Arthur",
            u"surname": u"of Camelot",
            u"extra": {
                u"quest": u"Grail",
                u"government": u"monarchy",
            },
        })
        contact_key = contact[u"key"]

        stored_contact = yield self.get_contact(api, contact_key)
        self.assert_contact(stored_contact, contact)

    @inlineCallbacks
    def test_create_invalid_fields(self):
        api = self.mk_api()
        (code, data) = yield self.request(
            api, "POST", "/contacts/", json.dumps({
                u"unknown_field": u"foo",
                u"not_the_field": u"bar",
            }))
        self.assertEqual(code, 400)
        self.assertEqual(data, {
            u"status_code": 400,
            u"reason": u"Invalid contact fields: not_the_field, unknown_field",
        })

    @inlineCallbacks
    def test_update(self):
        api = self.mk_api()
        contact = yield self.create_contact(
            api, name=u"Bob", msisdn=u"+12345")

        contact_key = contact[u"key"]
        resp = yield self.request(
            api, "PUT", "/contacts/" + contact_key, json.dumps({
                "msisdn": u"+6789",
            }))
        self.assert_contact_response(resp, {
            u"key": contact_key,
            u"created_at": contact[u"created_at"],
            u"msisdn": u"+6789",
            u"name": u"Bob",
        })

    @inlineCallbacks
    def test_update_extras(self):
        api = self.mk_api()
        contact = yield self.create_contact(
            api, name=u"Arthur", msisdn=u"+12345")

        contact_key = contact[u"key"]
        resp = yield self.request(
            api, "PUT", "/contacts/" + contact_key, json.dumps({
                u"msisdn": u"+6789",
                u"extra": {
                    u"quest": u"Grail",
                    u"government": u"monarchy",
                },
            }))
        self.assert_contact_response(resp, {
            u"key": contact_key,
            u"created_at": contact[u"created_at"],
            u"msisdn": u"+6789",
            u"name": u"Arthur",
            u"extra": {
                u"quest": u"Grail",
                u"government": u"monarchy",
            },
        })

        resp = yield self.request(
            api, "PUT", "/contacts/" + contact_key, json.dumps({
                u"extra": {
                    u"government": u"repressive",
                    u"assistant": u"Patsy",
                },
            }))
        self.assert_contact_response(resp, {
            u"key": contact_key,
            u"created_at": contact[u"created_at"],
            u"msisdn": u"+6789",
            u"name": u"Arthur",
            u"extra": {
                u"government": u"repressive",
                u"assistant": u"Patsy",
            },
        })

    @inlineCallbacks
    def test_update_non_existent_contact(self):
        api = self.mk_api()
        resp = yield self.request(
            api, "PUT", "/contacts/bad-id", json.dumps({
                "msisdn": u"+6789",
            }))
        self.assertEqual(resp, (404, {
            u"status_code": 404,
            u"reason": u"Contact 'bad-id' not found.",
        }))

    @inlineCallbacks
    def test_update_invalid_fields(self):
        api = self.mk_api()
        contact = yield self.create_contact(
            api, name=u"Bob", msisdn=u"+12345")

        contact_key = contact[u"key"]
        resp = yield self.request(
            api, "PUT", "/contacts/" + contact_key, json.dumps({
                u"unknown_field": u"foo",
                u"not_the_field": u"bar",
            }))
        self.assertEqual(resp, (400, {
            u"status_code": 400,
            u"reason": u"Invalid contact fields: not_the_field, unknown_field",
        }))

    @inlineCallbacks
    def test_delete(self):
        api = self.mk_api()
        contact = yield self.create_contact(
            api, name=u"Bob", msisdn=u"+12345")

        contact_key = contact[u"key"]
        resp = yield self.request(
            api, "DELETE", "/contacts/" + contact_key)
        self.assert_contact_response(resp, contact)

        exists = yield self.contact_exists(api, contact_key)
        self.assertFalse(exists)

    @inlineCallbacks
    def test_delete_non_existent_contact(self):
        api = self.mk_api()
        resp = yield self.request(api, "DELETE", "/contacts/bad-id")
        self.assertEqual(resp, (404, {
            u"status_code": 404,
            u"reason": u"Contact 'bad-id' not found.",
        }))


class TestContactsApi(VumiTestCase, ContactsApiTestMixin):
    def setUp(self):
        self.persistence_helper = self.add_helper(
            PersistenceHelper(use_riak=True, is_sync=False))

    def mk_config(self, config_dict):
        tempfile = self.mktemp()
        with open(tempfile, 'wb') as fp:
            yaml.safe_dump(config_dict, fp)
        return tempfile

    def mk_api(self):
        configfile = self.mk_config({
            "riak_manager": {
                "bucket_prefix": "test",
            },
        })
        return ContactsApi(configfile)

    @inlineCallbacks
    def request(self, api, method, path, body=None, headers=None, auth=True):
        if headers is None:
            headers = {}
        if auth:
            headers["X-Owner-ID"] = self.OWNER_ID.encode("utf-8")
        app_helper = AppHelper(app=api)
        resp = yield app_helper.request(
            method, path, data=body, headers=headers)
        data = yield resp.json()
        returnValue((resp.code, data))

    def _store(self, api):
        owner = self.OWNER_ID.encode("utf-8")
        return api.backend.get_contact_collection(owner).contact_store

    @inlineCallbacks
    def create_contact(self, api, **contact_data):
        contact = yield self._store(api).new_contact(**contact_data)
        returnValue(contact_to_dict(contact))

    @inlineCallbacks
    def get_contact(self, api, contact_key):
        contact = yield self._store(api).get_contact_by_key(contact_key)
        returnValue(contact_to_dict(contact))

    @inlineCallbacks
    def contact_exists(self, api, contact_key):
        from go.vumitools.contact.models import ContactNotFoundError
        try:
            yield self.get_contact(api, contact_key)
        except ContactNotFoundError:
            returnValue(False)
        else:
            returnValue(True)

    def test_init(self):
        configfile = self.mk_config({
            "riak_manager": {
                "bucket_prefix": "test",
            },
        })
        api = ContactsApi(configfile)
        self.assertTrue(isinstance(api.backend, RiakContactsBackend))
        self.assertTrue(isinstance(api.backend.riak_manager, TxRiakManager))

    def test_init_no_configfile(self):
        err = self.assertRaises(ValueError, ContactsApi)
        self.assertEqual(
            str(err),
            "Please specify a config file using --appopts=<config.yaml>")

    def test_init_no_riak_config(self):
        configfile = self.mk_config({})
        err = self.assertRaises(ValueError, ContactsApi, configfile)
        self.assertEqual(
            str(err),
            "Config file must contain a riak_manager entry.")

    def test_collections(self):
        api = self.mk_api()
        self.assertEqual(api.collections, (
            ('/contacts', api.backend.get_contact_collection),
        ))

    @inlineCallbacks
    def test_route(self):
        api = self.mk_api()
        collection = api.backend.get_contact_collection(
            self.OWNER_ID.encode("utf-8"))
        key, data = yield collection.create(None, {
            "msisdn": u"+12345",
        })
        contact = yield collection.contact_store.get_contact_by_key(key)
        code, data = yield self.request(
            api, "GET", '/contacts/%s' % (key,))
        self.assertEqual(code, 200)
        self.assertEqual(data, contact_to_dict(contact))


class TestFakeContactsApi(VumiTestCase, ContactsApiTestMixin):
    def setUp(self):
        try:
            from fake_go_contacts import Request, FakeContactsApi
        except ImportError as err:
            if "fake_go_contacts" not in err.args[0]:
                raise
            raise ImportError(" ".join([
                err.args[0],
                "(install from pypi or the 'verified-fake' directory)"]))

        self.req_class = Request
        self.api_class = FakeContactsApi

    def mk_api(self):
        return self.api_class("", "token-1", {})

    def request(self, api, method, path, body=None, headers=None, auth=True):
        if headers is None:
            headers = {}
        if auth:
            headers["Authorization"] = "Bearer token-1"
        resp = api.handle_request(self.req_class(
            method, path, body=body, headers=headers))
        return resp.code, json.loads(resp.body)

    def create_contact(self, api, **contact_data):
        return api.create_contact(contact_data)

    def get_contact(self, api, contact_key):
        return api.get_contact(contact_key)

    def contact_exists(self, api, contact_key):
        from fake_go_contacts import FakeContactsError
        try:
            self.get_contact(api, contact_key)
        except FakeContactsError:
            return False
        else:
            return True



"""
Cyclone application for Vumi Go contacts API.
"""

from vumi.persist.txriak_manager import TxRiakManager

from go_api.cyclone.handlers import ApiApplication
from go_contacts.backends.riak import RiakContactsBackend


class ContactsApi(ApiApplication):
    """
    :param IContactsBackend backend:
        A backend that provides a contact collection factory.
    """

    config_required = True

    def initialize(self, settings, config):
        self.backend = self._setup_backend(config)

    def _setup_backend(self, config):
        if "riak_manager" not in config:
            raise ValueError(
                "Config file must contain a riak_manager entry.")
        riak_manager = TxRiakManager.from_config(config['riak_manager'])
        backend = RiakContactsBackend(riak_manager)
        return backend

    @property
    def collections(self):
        return (
            ('/contacts', self.backend.get_contact_collection),
        )



"""
Tests for riak contacts backend and collection.
"""

from datetime import datetime

from twisted.internet.defer import inlineCallbacks, returnValue
from zope.interface.verify import verifyObject

from vumi.tests.helpers import VumiTestCase
from vumi.tests.helpers import PersistenceHelper

from go.vumitools.contact import ContactStore, ContactNotFoundError

from go_api.collections import ICollection
from go_api.collections.errors import (
    CollectionObjectNotFound, CollectionUsageError)

from go_contacts.backends.riak import (
    RiakContactsBackend, RiakContactsCollection)


class TestRiakContactsBackend(VumiTestCase):
    def setUp(self):
        self.persistence_helper = self.add_helper(
            PersistenceHelper(use_riak=True, is_sync=False))

    @inlineCallbacks
    def mk_backend(self):
        manager = yield self.persistence_helper.get_riak_manager()
        backend = RiakContactsBackend(manager)
        returnValue(backend)

    @inlineCallbacks
    def test_get_contacts_collection(self):
        backend = yield self.mk_backend()
        collection = backend.get_contact_collection("owner-1")
        self.assertEqual(collection.contact_store.user_account_key, "owner-1")
        self.assertTrue(isinstance(collection, RiakContactsCollection))


class TestRiakContactsCollection(VumiTestCase):
    def setUp(self):
        self.persistence_helper = self.add_helper(
            PersistenceHelper(use_riak=True, is_sync=False))

    @inlineCallbacks
    def mk_collection(self, owner_id):
        manager = yield self.persistence_helper.get_riak_manager()
        contact_store = ContactStore(manager, owner_id)
        collection = RiakContactsCollection(contact_store)
        returnValue(collection)

    EXPECTED_DATE_FORMAT = "%Y-%m-%d %H:%M:%S.%f"

    CONTACT_FIELD_DEFAULTS = {
        u'$VERSION': 2,
        u'bbm_pin': None,
        u'dob': None,
        u'email_address': None,
        u'facebook_id': None,
        u'groups': [],
        u'gtalk_id': None,
        u'mxit_id': None,
        u'name': None,
        u'surname': None,
        u'twitter_handle': None,
        u'wechat_id': None,
        u'extra': {},
        u'subscription': {},
    }

    def assert_contact(self, contact, expected_partial):
        expected = self.CONTACT_FIELD_DEFAULTS.copy()
        expected.update(expected_partial)
        if isinstance(expected.get("created_at"), datetime):
            expected["created_at"] = expected["created_at"].strftime(
                self.EXPECTED_DATE_FORMAT)
        self.assertEqual(contact, expected)

    def test_pick_fields(self):
        pick_fields = RiakContactsCollection._pick_fields
        self.assertEqual(
            pick_fields({"a": "1", "b": "2"}, ["a", "c"]),
            {"a": "1"})

    def test_pick_contact_fields(self):
        pick_contact_fields = RiakContactsCollection._pick_contact_fields
        self.assertEqual(
            pick_contact_fields({"msisdn": "+12345", "notfield": "xyz"}),
            {"msisdn": "+12345"})

    def test_check_contact_fields_success(self):
        check_contact_fields = RiakContactsCollection._check_contact_fields
        self.assertEqual(
            check_contact_fields({"msisdn": "+12345"}),
            {"msisdn": "+12345"})

    def test_check_contact_fields_raises(self):
        check_contact_fields = RiakContactsCollection._check_contact_fields
        err = self.assertRaises(
            CollectionUsageError, check_contact_fields,
            {"msisdn": "+12345", "notfield": "xyz"})
        self.assertEqual(str(err), "Invalid contact fields: notfield")

    def test_check_contact_fields_raises_multiple_fields(self):
        check_contact_fields = RiakContactsCollection._check_contact_fields
        err = self.assertRaises(
            CollectionUsageError, check_contact_fields,
            {"msisdn": "+12345", "notfield": "xyz", "badfield": "foo"})
        self.assertEqual(
            str(err), "Invalid contact fields: badfield, notfield")

    @inlineCallbacks
    def test_collection_provides_ICollection(self):
        """
        The return value of .get_row_collection() is an object that provides
        ICollection.
        """
        collection = yield self.mk_collection("owner-1")
        verifyObject(ICollection, collection)

    @inlineCallbacks
    def test_init(self):
        collection = yield self.mk_collection("owner-1")
        self.assertEqual(collection.contact_store.user_account_key, "owner-1")

    @inlineCallbacks
    def test_get(self):
        collection = yield self.mk_collection("owner-1")
        new_contact = yield collection.contact_store.new_contact(
            name=u"Bob", msisdn=u"+12345")
        contact = yield collection.get(new_contact.key)
        self.assert_contact(contact, {
            u'key': new_contact.key,
            u'created_at': new_contact.created_at,
            u'msisdn': u'+12345',
            u'name': u'Bob',
            u'user_account': u'owner-1',
        })

    @inlineCallbacks
    def test_get_non_existent_contact(self):
        collection = yield self.mk_collection("owner-1")
        d = collection.get("bad-contact-id")
        err = yield self.failUnlessFailure(d, CollectionObjectNotFound)
        self.assertEqual(str(err), "Contact 'bad-contact-id' not found.")

    @inlineCallbacks
    def test_create(self):
        collection = yield self.mk_collection("owner-1")
        key, contact = yield collection.create(None, {
            "msisdn": u"+12345",
            "name": u"Arthur",
            "surname": u"of Camelot",
        })
        new_contact = yield collection.contact_store.get_contact_by_key(key)
        self.assert_contact(contact, {
            u'key': new_contact.key,
            u'created_at': new_contact.created_at,
            u'msisdn': u'+12345',
            u'name': u'Arthur',
            u'surname': u'of Camelot',
            u'user_account': u'owner-1',
        })
        self.assertEqual(new_contact.key, key)
        self.assertEqual(new_contact.name, u"Arthur")
        self.assertEqual(new_contact.surname, u"of Camelot")
        self.assertEqual(new_contact.msisdn, u"+12345")

    @inlineCallbacks
    def test_create_with_id_fails(self):
        collection = yield self.mk_collection("owner-1")
        d = collection.create(u"foo", {
            "msisdn": u"+12345",
            "name": u"Sir Gawain",
        })
        err = yield self.failUnlessFailure(d, CollectionUsageError)
        self.assertEqual(
            str(err), "A contact key may not be specified in contact creation")

    @inlineCallbacks
    def test_create_invalid_fields(self):
        collection = yield self.mk_collection("owner-1")
        d = collection.create(None, {
            "unknown_field": u"foo",
            "not_the_field": u"bar",
        })
        err = yield self.failUnlessFailure(d, CollectionUsageError)
        self.assertEqual(
            str(err), "Invalid contact fields: not_the_field, unknown_field")

    @inlineCallbacks
    def test_create_invalid_field_value(self):
        collection = yield self.mk_collection("owner-1")
        d = collection.create(None, {
            "msisdn": 5,
        })
        err = yield self.failUnlessFailure(d, CollectionUsageError)
        self.assertEqual(
            str(err), "Value 5 is not a unicode string.")

    @inlineCallbacks
    def test_update(self):
        collection = yield self.mk_collection("owner-1")
        new_contact = yield collection.contact_store.new_contact(
            name=u"Bob", msisdn=u"+12345")
        contact = yield collection.update(new_contact.key, {
            "msisdn": u"+6789",
        })
        self.assert_contact(contact, {
            u'key': new_contact.key,
            u'created_at': new_contact.created_at,
            u'msisdn': u"+6789",
            u'name': u'Bob',
            u'user_account': u'owner-1',
        })

    @inlineCallbacks
    def test_update_non_existent_contact(self):
        collection = yield self.mk_collection("owner-1")
        d = collection.update("bad-contact-id", {})
        err = yield self.failUnlessFailure(d, CollectionObjectNotFound)
        self.assertEqual(str(err), "Contact 'bad-contact-id' not found.")

    @inlineCallbacks
    def test_update_invalid_fields(self):
        collection = yield self.mk_collection("owner-1")
        new_contact = yield collection.contact_store.new_contact(
            name=u"Bob", msisdn=u"+12345")
        d = collection.update(new_contact.key, {
            "unknown_field": u"foo",
            "not_the_field": u"bar",
        })
        err = yield self.failUnlessFailure(d, CollectionUsageError)
        self.assertEqual(
            str(err), "Invalid contact fields: not_the_field, unknown_field")

    @inlineCallbacks
    def test_update_invalid_field_value(self):
        collection = yield self.mk_collection("owner-1")
        new_contact = yield collection.contact_store.new_contact(
            name=u"Bob", msisdn=u"+12345")
        d = collection.update(new_contact.key, {
            "msisdn": None,
        })
        err = yield self.failUnlessFailure(d, CollectionUsageError)
        self.assertEqual(
            str(err), "None is not allowed as a value for non-null fields.")

    @inlineCallbacks
    def test_delete(self):
        collection = yield self.mk_collection("owner-1")
        new_contact = yield collection.contact_store.new_contact(
            name=u"Bob", msisdn=u"+12345")
        contact = yield collection.delete(new_contact.key)
        self.assert_contact(contact, {
            u'key': new_contact.key,
            u'created_at': new_contact.created_at,
            u'msisdn': u'+12345',
            u'name': u'Bob',
            u'user_account': u'owner-1',
        })
        d = collection.contact_store.get_contact_by_key("owner-1")
        yield self.failUnlessFailure(d, ContactNotFoundError)

    @inlineCallbacks
    def test_delete_non_existent_contact(self):
        collection = yield self.mk_collection("owner-1")
        d = collection.delete("bad-contact-id")
        err = yield self.failUnlessFailure(d, CollectionObjectNotFound)
        self.assertEqual(str(err), "Contact 'bad-contact-id' not found.")



"""
Riak contacts backend and collection.
"""

from twisted.internet.defer import inlineCallbacks, returnValue
from zope.interface import implementer

from vumi.persist.fields import ValidationError
from go.vumitools.contact import (
    ContactStore, ContactNotFoundError, Contact)

from go_api.collections import ICollection
from go_api.collections.errors import (
    CollectionObjectNotFound, CollectionUsageError)


def contact_to_dict(contact):
    """
    Turn a contact into a dict we can return.

    TODO: Move this into the contact model in vumi-go or something.
    """
    contact_dict = {
        u"extra": dict(contact.extra),
        u"subscription": dict(contact.subscription),
    }
    for key, value in contact.get_data().iteritems():
        if key.startswith(u"extras-") or key.startswith(u"subscription-"):
            # We already have these.
            continue
        contact_dict[key] = value
    return contact_dict


class RiakContactsBackend(object):
    def __init__(self, riak_manager):
        self.riak_manager = riak_manager

    def get_contact_collection(self, owner_id):
        contact_store = ContactStore(self.riak_manager, owner_id)
        return RiakContactsCollection(contact_store)


@implementer(ICollection)
class RiakContactsCollection(object):
    def __init__(self, contact_store):
        self.contact_store = contact_store

    @staticmethod
    def _pick_fields(data, keys):
        """
        Return a sub-dictionary of all the items from ``data`` whose
        keys are listed in ``keys``.
        """
        return dict((k, data[k]) for k in keys if k in data)

    @classmethod
    def _pick_contact_fields(cls, data):
        """
        Return a sub-dictionary of the items from ``data`` that are valid
        contact fields.
        """
        return cls._pick_fields(data, Contact.field_descriptors.keys())

    @classmethod
    def _check_contact_fields(cls, data):
        """
        Return a sub-dictionary of the items from ``data`` that are valid
        and raise a :class:`CollectionUsageError` if any fields are not.
        """
        fields = cls._pick_contact_fields(data)
        given_keys = set(data.keys())
        valid_keys = set(fields.keys())
        if given_keys != valid_keys:
            raise CollectionUsageError(
                "Invalid contact fields: %s" % ", ".join(
                    sorted(given_keys - valid_keys)))
        return fields

    def all_keys(self):
        """
        Return an iterable over all keys in the collection. May return a
        deferred instead of the iterable.
        """
        raise NotImplementedError()

    def all(self):
        """
        Return an iterable over all objects in the collection. The iterable may
        contain deferreds instead of objects. May return a deferred instead of
        the iterable.
        """
        raise NotImplementedError()

    @inlineCallbacks
    def get(self, object_id):
        """
        Return a single object from the collection. May return a deferred
        instead of the object.
        """
        try:
            contact = yield self.contact_store.get_contact_by_key(object_id)
        except ContactNotFoundError:
            raise CollectionObjectNotFound(object_id, "Contact")
        returnValue(contact_to_dict(contact))

    @inlineCallbacks
    def create(self, object_id, data):
        """
        Create an object within the collection. May return a deferred.

        If ``object_id`` is ``None``, an identifier will be generated.
        """
        fields = self._check_contact_fields(data)
        if object_id is not None:
            raise CollectionUsageError(
                "A contact key may not be specified in contact creation")
        try:
            contact = yield self.contact_store.new_contact(**fields)
        except ValidationError, e:
            raise CollectionUsageError(str(e))
        returnValue((contact.key, contact_to_dict(contact)))

    @inlineCallbacks
    def update(self, object_id, data):
        """
        Update an object. May return a deferred.

        ``object_id`` may not be ``None``.
        """
        fields = self._check_contact_fields(data)
        try:
            contact = yield self.contact_store.update_contact(
                object_id, **fields)
        except ContactNotFoundError:
            raise CollectionObjectNotFound(object_id, "Contact")
        except ValidationError, e:
            raise CollectionUsageError(str(e))
        returnValue(contact_to_dict(contact))

    @inlineCallbacks
    def delete(self, object_id):
        """
        Delete an object. May return a deferred.
        """
        try:
            contact = yield self.contact_store.get_contact_by_key(object_id)
        except ContactNotFoundError:
            raise CollectionObjectNotFound(object_id, "Contact")
        contact_data = contact_to_dict(contact)
        yield contact.delete()
        returnValue(contact_data)
